import json

from traceback import format_exc
from functools import partial
from sqlor.dbpools import DBPools
from appPublic.streamhttpclient import StreamHttpClient, liner
from appPublic.dictObject import DictObject
from appPublic.myTE import MyTemplateEngine
from appPublic.log import debug, exception, error
from ahserver.globalEnv import password_decode
from ahserver.serverenv import get_serverenv

def get_dbname():
	f = get_serverenv('get_module_dbname')
	dbname = f('uapi')
	return dbname

class UAPI:
	def __init__(self, env={}):
		self.te = MyTemplateEngine([], env=env)
		self.env = env
		self.auth_api = None
		self.auth_ret = None
	
	async def rendertmpl(self, tmplstr, params={}):
		if tmplstr is None:
			return None
		ns = self.env.copy()
		ns.update(params)
		te = MyTemplateEngine([], env=self.env)
		return te.renders(tmplstr, ns)

	async def cvt_upappapi2uapi(self, upappapiid, callerid, params={}):
		self.env.update(params)
		db = DBPools()
		dbname = get_dbname()
		uapi = None
		auth_uapi = None
		async with db.sqlorContext(dbname) as sor:
			upappapi = await self.get_upappapis(sor, upappapiid)
			kinfo = await self.get_orgapikey(sor, upappapi.upappid, callerid)
			self.env.update(kinfo)
			auth_uapi = None
			if kinfo.auth_apiid:
				auth_upapi = await self.get_upappapis(sor, kinfo.auth_apiid)
				auth_uapi = await self.get_uapi(sor, kinfo.auth_upapi.uapiid)
			uapi = await self.get_uapi(sor, upappapi.uapiid)
			return auth_uapi, uapi
		return None, None

	async def __call__(self, upapiid, callerid, params={}):
		"""
		"""
		auth_uapi, uapi = await self.cvt_upappapi2uapi(upapiid, 
										callerid, params=params)
		if uapi is None:
			return
		if auth_uapi:
			await self.do_auth(auth_uapi)
		async for line in self.stream_gen(uapi):
				yield line

	async def request(self, upapiid, callerid, params={}):
		auth_uapi, uapi = await self.cvt_upappapi2uapi(upapiid, 
										callerid, params=params)
		if auth_uapi:
			await self.do_auth(auth_uapi)
		return await self.do_call(uapi)

	async def do_auth(self, auth_uapi):
		b = await self.do_call(auth_uapi)
		d = json.loads(b.encode('utf-8'))
		self.env.update(d)
		return
	
	async def do_call(self, api, params={}):
		url = self.env.get('baseurl') + api.path
		method = api.httpmethod
		header = await self.rendertmpl(api.headers)
		header = json.loads(headers)
		body = await self.rendertmpl(api.body)
		_params = await self.rendertmpl(api.params)
		if _params:
			_params = json.loads(_params)
		shc = StreamHttpClient()
		b = await shc.request(method, url, 
							headers=headers,
							data=body,
							params=_params)

		d = json.loads(b.encode('utf-8'))
		if api.response:
			return await self.rendertmpl(api.response, params=d)
		return d

	async def stream_gen(self, api, params={}):
		url = self.env.get('baseurl') + api.path
		method = api.httpmethod
		headers = await self.rendertmpl(api.headers)
		headers = json.loads(headers)
		body = await self.rendertmpl(api.body)
		if body:
			bdy = json.loads(body)
			bdy['stream'] = True
			body = bdy.dumps(bdy, ensure_ascii=False)
		_params = await self.rendertmpl(api.params)
		if _params:
			_params = json.loads(_params)
		shc = StreamHttpClient()
		gen = shc(method, url, 
						headers=headers,
						data=body,
						params=_params)
		chunk_match = api.chunk_match or ''
		cmlen = len(chunk_match)
		async for line in liner(gen):
			if line.startswith(chunk_match):
				cvt_line = self.streamline_handle(line[cmlen:], 
												api.streamresponse)
				if cvt_line is not None:
					yield cvt_line

	async def streamline_handle(self, line, resptmpl):
		try:
			dic = json.loads(line)
			if resptmpl:
				jstr = await self.rendertmpl(resptmpl, params=dic)
				jstr += '\n'
			else:
				jstr = json.dumps(self.resp_json, ensure_ascii=False) + '\n'
			return jstr
		except Exception as e:
			exception(f'{line=}\n{e=},{format_exc()}')
			return None
			
	async def get_upappapis(self, sor, upapiid):
		recs = await sor.R('upappapis', {'id': upapiid})
		if len(recs) < 1:
			e = Exception(f'{upapiid} not found in table(upappapis)')
			exception(f'{e}\n{format_exc()}')
			raise e
		return recs[0]
	
	async def get_uapi(self, sor, uapiid):
		recs = await sor.R('uapi', {'id': uapiid})
		if len(recs) < 1:
			e = Exception(f'{uapiid} not found in table(uapi)')
			exception(f'{e}\n{format_exc()}')
			raise e
		return recs[0]
		
	async def get_orgapikey(self, sor, upappid, callerid):
		"""
		argumemts:
			upappid: upappid which will make call to
			orgid: owner organization or user which as the caller of the call
		return:
			None: this orgid has not gotton apikey from upapp
			dict if apikey and upapp infos
		"""
		sql = """select 
a.myappid, 
a.apisetid, 
a.auth_apiid,
a.baseurl,
b.*
from upapp a, upappkey b 
where a.id = b.upappid
	and a.id = ${appid}$
	and b.ownerid = ${orgid}$"""
		recs = await sor.sqlExe(sql, {'appid':upappid, 'orgid': callerid})
		if len(recs) < 1:
			e = Exception(f'{appid=}, {callerid=} has not apikey')
			exception(f'{e}, {format_exc()}')
			raise e
		r = recs[0]
		return DictObject(**{
			'apikey':r.apikey,
			'secretkey':r.secretkey,
			'apisetid': r.apisetid,
			'auth_apiid': r.auth_apiid,
			'myappid': r.myappid
		})
	
	async def get_authapi(self, sor):
		sql = "select * from jsonhttpapi where name = 'auth' and appid=${appid}$"
		recs = await sor.sqlExe(sql, {'appid': self.appid})
		if len(recs):
			return recs[0]
		return None

	async def get_apiinfo(self, orgids, apiname):
		db = DBPools()
		dbname = get_serverenv('get_module_dbname')('uapi')
		async with db.sqlorContext(dbname) as sor:
			sql = """select c.baseurl,
c.name as appname,
b.*,
a.apikey,
a.secretkey
from upappkey a, jsonhttpapi b, upapp c
where
	a.upappid=b.appid
	and a.upappid = c.id
	and b.appid = ${appid}$
	and b.name = ${apiname}$
	and a.ownerid = ${orgid}$
"""
			for orgid in orgids:
				ns = {
					"orgid": orgid,
					"apiname": appname,
					"appid": self.appid
				}
				recs = await sor.sqlExe(sql, ns)
				if len(recs) > 0:
					rec = recs[0]
					rec.apikey = password_decode(rec.apikey)
					if rec.secretkey:
						rec.secretkey = password_decode(rec.secretkey)
					if rec.need_auth and self.auth_api is None:
						self.auth_api = await self.get_authapi(sor)
					if rec.need_auth and self.auth_api is None:
						e = Exception(f'{rec.appname} app has not auth api but need')
						exception(f'{e=}, {format_exc()}')
						raise e
					rec.auth_api = self.auth_api
					rec.apikey_orgid = orgid
					return recs[0]
		return None

